#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import urllib.parse
import smtplib

class MailDropoff():
	class Scheme(enum.Enum):
		LMTP = "lmtp"
		LMTP_STARTTLS = "lmtp+starttls"
		SMTP = "smtp"
		SMTPS = "smtps"
		SMTP_STARTTLS = "smtp+startls"

	_SCHEME_BY_NAME = { scheme.value: scheme for scheme in Scheme }

	def __init__(self, scheme: Scheme, host: str, port: int | None = None, username: str | None = None, password: str | None = None):
		self._scheme = scheme
		self._host = host
		if port is None:
			self._port = {
				self.Scheme.LMTP:			24,
				self.Scheme.LMTP_STARTTLS:	24,
				self.Scheme.SMTP:			25,
				self.Scheme.SMTPS:			465,
				self.Scheme.SMTP_STARTTLS:	25,
			}[self._scheme]
		else:
			self._port = port
		self._username = username
		self._password = password

	@classmethod
	def parse_uri(cls, uri: str):
		parsed = urllib.parse.urlparse(uri)
		if parsed.scheme not in cls._SCHEME_BY_NAME:
			raise ValueError(f"\"{parsed.scheme}\" is not a valid URI scheme, supported are: {', '.join(sorted(cls._SCHEME_BY_NAME))}")
		scheme = cls._SCHEME_BY_NAME[parsed.scheme]
		if ":" in parsed.netloc:
			(host, port) = parsed.netloc.split(":", maxsplit = 1)
			port = int(port)
		else:
			(host, port) = (parsed.netloc, None)
		return cls(scheme = scheme, host = host, port = port)

	@property
	def username(self):
		return self._username

	@username.setter
	def username(self, value: str):
		self._username = value

	@property
	def password(self):
		return self._password

	@password.setter
	def password(self, value: str):
		self._password = value

	def postall(self, mails: list["Email"]):
		conn_class = smtplib.SMTP_SSL if (self._scheme == self.Scheme.SMTPS) else smtplib.SMTP
		with conn_class(self._host, self._port) as conn:
			try:
				if self._scheme == self.Scheme.SMTP_STARTTLS:
					conn.starttls()
				if (self._username is not None) and (self._password is not None):
					conn.login(self._username, self._password)

				for mail in mails:
					serialized_mail = mail.serialize()
					conn.send_message(serialized_mail.content, to_addrs = serialized_mail.recipients)
			finally:
				conn.quit()

	def post(self, mail: "Email"):
		return self.postall([ mail ])

	def __str__(self):
		return f"MailDropoff<{self._scheme.value}: {self._host}:{self._port}>"



#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
import getpass
import mailcoil
from mailcoil.FriendlyArgumentParser import FriendlyArgumentParser

class CLIMailer():
	def __init__(self, args):
		self._args = args

	def _get_username_password(self):
		if self._args.auth_username is None:
			return (None, None)

		if self._args.auth_password is None:
			password = getpass.getpass()
		else:
			with open(self._args.auth_password) as f:
				password = f.read().rstrip("\r\n")
		return (self._args.auth_username, password)

	def run(self):
		mail = mailcoil.Email(from_address = getattr(self._args, "from"), wrap_text = self._args.wrap_text)
		if self._args.subject:
			mail.subject = self._args.subject

		mail.to(*self._args.to)
		mail.cc(*self._args.cc)
		mail.bcc(*self._args.bcc)

		if self._args.text is not None:
			with open(self._args.text) as f:
				mail.text = f.read()

		if self._args.html is not None:
			with open(self._args.html) as f:
				mail.html = f.read()

		for filename in self._args.attach_file:
			mail.attach(filename)
		for filename in self._args.attach_file_inline:
			mail.attach(filename, inline = True)

		if (self._args.smime_sign is not None) or (len(self._args.smime_encrypt) > 0):
			mail.security = mailcoil.CMS()
			if (self._args.smime_sign is not None):
				mail.security.sign(signer_certfile = self._args.smime_sign["cert"], signer_keyfile = self._args.smime_sign["key"], ca_certfile = self._args.smime_sign.get("ca"))
			mail.security.encrypt(*self._args.smime_encrypt)

		if self._args.smtp_server_uri is None:
			print(mail.serialize().content)
		else:
			dropoff = mailcoil.MailDropoff.parse_uri(self._args.smtp_server_uri)
			(username, password) = self._get_username_password()
			if username is not None:
				dropoff.username = username
				dropoff.password = password
			dropoff.post(mail)

def main():
	def sign_params(text):
		split_text = text.split(":")
		if len(split_text) == 2:
			return { "cert": split_text[0], "key": split_text[1] }
		elif len(split_text) == 3:
			return { "cert": split_text[0], "key": split_text[1], "ca": split_text[2] }
		else:
			raise ValueError(f"Only two or three arguments supported: {text}")

	parser = FriendlyArgumentParser(description = f"Simple command line tool to demonstrate Mailcoil v{mailcoil.VERSION}")
	parser.add_argument("-u", "--smtp-server-uri", metavar = "uri", help = "SMTP server dropoff in URI format. E.g., 'smtp://127.0.0.1', or 'smtps://mail.gmx.net:1122'. If omitted, just prints the mail on the command line.")
	parser.add_argument("-U", "--auth-username", metavar = "username", help = "Authenticate against SMTP server using this username.")
	parser.add_argument("--auth-password", metavar = "filename", help = "When using authentication, read the password from this file. Otherwise, password is requested on the command line.")
	parser.add_argument("-f", "--from", metavar = "email", required = True, help = "From address. Mandatory argument.")
	parser.add_argument("-s", "--subject", metavar = "text", help = "Text to use as subject.")
	parser.add_argument("-t", "--to", metavar = "email", action = "append", default = [ ], help = "Send email to this address in the 'To' field. May be specified multiple times.")
	parser.add_argument("-c", "--cc", metavar = "email", action = "append", default = [ ], help = "Send email to this address in the 'CC' field. May be specified multiple times.")
	parser.add_argument("-b", "--bcc", metavar = "email", action = "append", default = [ ], help = "Send email to this address in the 'BCC' field. May be specified multiple times.")
	parser.add_argument("-T", "--text", metavar = "filename", help = "Use the data in this text file as text content.")
	parser.add_argument("-H", "--html", metavar = "filename", help = "Use the data in this text file as HTML content.")
	parser.add_argument("--wrap-text", action = "store_true", help = "Wrap text content after 72 columns.")
	parser.add_argument("-a", "--attach-file", metavar = "filename", action = "append", default = [ ], help = "Attach this file to your email. May be specified multiple times.")
	parser.add_argument("--attach-file-inline", metavar = "filename", action = "append", default = [ ], help = "Attach this file to your email with Content-Disposition \"inline\". May be specified multiple times.")
	parser.add_argument("--smime-sign", type = sign_params, metavar = "cert:key[:ca_cert]", help = "Sign email using CMS (often referred to as \"S/MIME\"). Requires at least two filenames, one for the PEM-encoded key and the other for the cert.")
	parser.add_argument("--smime-encrypt", metavar = "cert", action = "append", default = [ ], help = "Encrypt email using CMS (often referred to as \"S/MIME\"). Requires the certificate file of the intended recipient. Can be specified multiple times to encrypt email for multiple recipients.")
	parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
	args = parser.parse_args(sys.argv[1:])

	cli_mailer = CLIMailer(args)
	cli_mailer.run()
	return 0

if __name__ == "__main__":
	sys.exit(main())



#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import email
import time
import textwrap
import dataclasses
import mimetypes
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
import mailcoil
from .Exceptions import NoRecipientException, NoBodyException

@dataclasses.dataclass
class MailAddress():
	mail: str
	name: str | None = None

	def encode(self):
		if self.name is not None:
			return email.utils.formataddr((self.name, self.mail))
		else:
			return self.mail

	@classmethod
	def parse(cls, addr: "MailAddress | str"):
		if isinstance(addr, cls):
			return addr
		else:
			((name, mail), ) = email.utils.getaddresses([ addr ])
			if name == "":
				return cls(mail = mail)
			else:
				return cls(mail = mail, name = name)

@dataclasses.dataclass
class SerializedEmail():
	recipients: list[str]
	content: bytes


class MIMEGeneric(MIMENonMultipart):
	def __init__(self, payload: bytes, maintype: str, subtype: str, encoder: "callable"):
		MIMENonMultipart.__init__(self, maintype, subtype, policy = None)
		self.set_payload(payload)
		encoder(self)

@dataclasses.dataclass(slots = True)
class Attachment():
	data: bytes
	maintype: str
	subtype: str
	filename: str
	inline: bool
	content_id: str

	def as_mime(self):
		mime = MIMEGeneric(self.data, self.maintype, self.subtype, encoder = email.encoders.encode_base64)
		mime["Content-Disposition"] = f"{'inline' if self.inline else 'attachment'}; filename=\"{self.filename}\""
		mime["Content-ID"] = self.content_id
		return mime

class Email():
	def __init__(self, from_address: MailAddress | str, subject: str | None = None, text: str | None = None, wrap_text: bool = False, html: str | None = None, security: "SMIME | None" = None):
		self._from = MailAddress.parse(from_address)
		self._to = [ ]
		self._cc = [ ]
		self._bcc = [ ]
		self._subject = subject
		self._text = text
		self._wrap_text = wrap_text
		self._html = html
		self._security = security
		self._datetime = time.time()
		self._message_id = f"<{os.urandom(16).hex()}@mailcoil>"
		self._attachments = [ ]

	@property
	def recipient_count(self):
		return len(self._to) + len(self._cc) + len(self._bcc)

	def to(self, *mail_addresses: tuple[MailAddress | str]):
		self._to += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	def cc(self, *mail_addresses: tuple[MailAddress | str]):
		self._cc += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	def bcc(self, *mail_addresses: tuple[MailAddress | str]):
		self._bcc += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	@property
	def subject(self):
		return self._subject

	@subject.setter
	def subject(self, value: str):
		self._subject = value

	@property
	def text(self):
		return self._text

	@text.setter
	def text(self, value: str):
		self._text = value

	@property
	def html(self):
		return self._html

	@html.setter
	def html(self, value: str):
		self._html = value

	@property
	def wrapped_text(self):
		if self._wrap_text:
			return self._wrap_paragraphs(self.text)
		else:
			return self.text

	@property
	def security(self):
		return self._security

	@security.setter
	def security(self, value: str):
		self._security = value

	def _mimetype(self, filename: str, override: str | None):
		if override is None:
			(mimetype, _) = mimetypes.guess_type(filename)
			if mimetype is None:
				return "application/octet-stream"
			else:
				return mimetype
		else:
			return override

	def attach_data(self, data: bytes, filename: str, mimetype: str | None = None, inline: bool = False, cid: str | None = None):
		(maintype, subtype) = self._mimetype(filename, mimetype).split("/")
		content_id = f"cid{len(self._attachments)}" if (cid is None) else cid
		attachment = Attachment(data = data, maintype = maintype, subtype = subtype, filename = filename, inline = inline, content_id = content_id)
		self._attachments.append(attachment)
		return f"cid:{attachment.content_id}"

	def attach(self, filename: str, mimetype: str | None = None, inline: bool = False, cid: str | None = None):
		with open(filename, "rb") as f:
			data = f.read()
		return self.attach_data(data, filename = os.path.basename(filename), mimetype = mimetype, inline = inline, cid = cid)

	@staticmethod
	def _wrap_paragraphs(text: str) -> str:
		wrapped = [ ]
		for paragraph in text.split("\n"):
			parwrapped = textwrap.wrap(paragraph, width = 72)
			if len(parwrapped) == 0:
				parwrapped = [ "" ]
			wrapped += parwrapped
		return "\n".join(wrapped)

	def _render_text_quopri(self, text: str, subtype: str):
		part = MIMEGeneric(text.encode("utf-8"), "text", subtype, encoder = email.encoders.encode_quopri)
		part.set_param("charset", "utf-8", header = "Content-Type")
		return part

	def _layer_text_content(self):
		if self.recipient_count == 0:
			raise NoRecipientException("Mail has no To, CC, or BCC set. Unable to serialize.")

		if (self.text is None) and (self.html is None):
			raise NoBodyException("Mail has no text or HTML content.")

		if (self.text is not None) and (self.html is not None):
			# text and HTML as multipart/alternative
			msg = MIMEMultipart("alternative")
			msg.attach(self._render_text_quopri(self.wrapped_text, "plain"))
			msg.attach(self._render_text_quopri(self.html, "html"))
		elif self.text is not None:
			msg = self._render_text_quopri(self.wrapped_text, "plain")
		else:
			msg = self._render_text_quopri(self.html, "html")
		return msg

	def _layer_attachments(self, prev_layer: "MIMEBase"):
		if len(self._attachments) == 0:
			msg = prev_layer
		else:
			msg = MIMEMultipart("related")
			msg.attach(prev_layer)
			for attachment in self._attachments:
				msg.attach(attachment.as_mime())
		return msg

	def _layer_security(self, prev_layer: "MIMEBase"):
		if self._security is None:
			msg = prev_layer
		else:
			msg = self._security.process(prev_layer)
		return msg

	def serialize(self):
		msg = self._layer_text_content()
		msg = self._layer_attachments(msg)
		msg = self._layer_security(msg)

		if self._subject is not None:
			msg["Subject"] = self._subject
		msg["Message-ID"] = self._message_id
		msg["Date"] = email.utils.formatdate(self._datetime, localtime = True)
		msg["User-Agent"] = f"mailcoil v{mailcoil.VERSION}"
		msg["From"] = self._from.encode()
		if len(self._to) > 0:
			msg["To"] = ", ".join([address.encode() for address in self._to ])
		if len(self._cc) > 0:
			msg["CC"] = ", ".join([address.encode() for address in self._cc ])
		if len(self._bcc) > 0:
			msg["BCC"] = ", ".join([address.encode() for address in self._bcc ])
		return SerializedEmail(content = msg, recipients = [ addr.mail for addr in (self._to + self._cc + self._bcc) ])
